"""
hash package written in pure python 3.
Copyright (C) 2021-2022 Vincenzo Palazzo vincenzopalazzodev@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
USA.
"""
from math import sqrt


class Utils:
    @staticmethod
    def generate_random_prime(target_number: int) -> int:
        """
        Trivial function to generate a random prime number,
        from a target number.
        :return: The prime number generated.
        """
        for number in range(target_number + 1, (2 * target_number) + 1):
            is_prime = True
            for prime in range(3, int(sqrt(number) - 1), -1):
                if prime % 2 == 0:
                    is_prime = False
                    break
            if is_prime:
                return number
        raise Exception(f"Prime number not found with target number {target_number}!")

    @staticmethod
    def convert_str_to_byte(input_str: str) -> str:
        """Convert the input string into the sequence of bytes into a string form"""
        return "".join(format(ord(i), "08b") for i in input_str)



"""
Graph data structure implementation with adjacency list in Python 3
Copyright (C) 2021-2022 Vincenzo Palazzo vincenzopalazzodev@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
USA.
"""
from abc import ABC, abstractmethod


class Node:
    """
    Node implementation that contains the value of the element.
    This class contains a list of children and the value of the node.
    """

    def __init__(self, value) -> None:
        self.value = value
        self.children = []

    def add_node(self, value) -> None:
        node = Node(value)
        self.children.append(node)


class Graph(ABC):
    """Graph interface"""

    @abstractmethod
    def add_edge(self, u, v, opts: dict = None):
        """Add the edge from A to b"""
        pass

    @abstractmethod
    def dfs(self) -> None:
        pass

    @abstractmethod
    def bfs(self) -> None:
        pass



from .sort import Sort, Order
from cpstl.datatype.heap import MaxHeap, MinHeap


class HeapSort(Sort):
    """Heap sort algorithm pure Python 3 implementation"""

    def sort(self, sequence: list, order=Order.INCREASE) -> list:
        heap = self.build_heap(sequence, order)
        print("---- start ------")
        offset = 1
        for idx in reversed(range(1, heap.len())):
            print(f"idx {idx}")
            heap.swap(0, idx)
            heap.heapify(0, size_offset=offset)
            offset += 1
        return heap.to_list()

    def build_heap(self, sequence: list, order: Order = Order.INCREASE):
        """"""
        if order == Order.INCREASE:
            return MaxHeap(sequence)
        return MinHeap(sequence)



"""
Similarity interface to implements similarity algorithms.
Copyright (C) 2021-2022 Vincenzo Palazzo vincenzopalazzodev@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
USA.
"""
from abc import ABC, abstractmethod


class ISimilarity(ABC):
    """
    Each similarity algorithm needs to implement this interface
    to make the similarity concept portable between concepts.

    author: Vincenzo Palazzo https://github.com/vincenzopalazzo
    """

    @abstractmethod
    def similarity(self, set_a: list, set_b: list) -> list:
        """
        Calculate the similarity between two sets and return
        a new set with all the similar elements.
        :param set_a: first set fo elements
        :param set_b: a second set of elements
        :return: a new set with a similar element between the two sets given in input
        """
        pass



"""
Flajolet–Martin (FM) Sketch algorithm in pure python 3.
Copyright (C) 2021-2022 Vincenzo Palazzo vincenzopalazzodev@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
USA.
"""
from ..hash.universal_hash import UniversalHash


class FMSketch(object):
    """
    Flajolet–Martin (FM) Sketch algorithm in pure python 3.

    The problem that we are going to solve with this algorithm is the
    following on:

    Let S be a multi-set of N integers, namely, two elements of S may be identical. Each integer is in
    the range of [0, D] where D is some polynomial of N. The distinct element counting problem is to
    find out exactly how many distinct elements there are in S. We will use F to denote the answer.
    For example, given S = {1, 5, 10, 5, 15, 1}, F = 4.

    Clearly, using O(N) words of space, the problem can be solved easily in O(N log N) time by
    sorting, or O(N) expected time with hashing. However, in many applications, the amount of space
    at our disposal can be much smaller.

    The following algorithm allowed us to count unique elements in only O(log N)
    bits. Hence, our goal is to obtain an approximate answer F˜ whose accuracy has a probabilistic
    guarantee.

    Pseudo Code-Stepwise Solution:

        1. Selecting a hash function h so each element in the set is mapped to a string to at least log2n bits;
        2. For each element x, r(x)= length of trailing zeroes in h(x);
        3. R= max(r(x));

    Now it is possible to estimate the distinct elements by the following formula
    Distinct elements= 2^R

    author: Vincenzo Palazzo https://github.com/vincenzopalazzo
    """

    @staticmethod
    def count_distinct(input_list: list, bits: int = 64) -> int:
        """
        Return the number of distinct elements in the input list.
        :return: The FM estimation of distinct elements in the list of elements.
        """
        max_zeros_sum = 0
        # number of univers! that in this case is the number of bits
        universal_hashing = UniversalHash(size=pow(2, bits))
        for index in range(len(input_list)):
            # The hash function returns an integer that is a binary presentation
            # of the hash value returned.
            hash_val = bin(universal_hashing.hash(input_list[index]))[2:]
            zeros_sum = 0
            for val in range(len(hash_val) - 1, 0, -1):
                if hash_val[val] == "0":
                    zeros_sum += 1
                else:
                    break
            if zeros_sum > max_zeros_sum:
                max_zeros_sum = zeros_sum
        return pow(2, max_zeros_sum)



"""Genetic implementation of the Heap that give the
possibility to implement Min and Max heap.

If you want refresh your mind on how the heap is implemented, it is
possible read some well known book introduction you can read this blog post
https://www.educative.io/blog/data-structure-heaps-guide"""
from abc import ABC, abstractmethod


class Heap(ABC):
    """Generic implementation of the heap that implement
    some common function and leave to the subclass redefine
    some behaivord"""

    def __init__(self, array: list = []) -> None:
        self.heap = array
        self.from_list(array)

    def from_list(self, array: []) -> bool:
        """Build the heap from the list"""
        self.heap = array
        for idx in range(self.len(), -1, -1):
            self.heapify(idx)
        return self.verify()

    @staticmethod
    def parent(idx: int) -> int:
        """Calculate the parent index"""
        return idx - 1 // 2

    @staticmethod
    def left(idx: int) -> int:
        return idx * 2 + 1

    @staticmethod
    def right(idx: int) -> int:
        return (idx * 2) + 2

    def insert(self, value) -> bool:
        """Insert element inside the heap, this operation
        swap the element from the bottom to the top, and can
        be costly, a solution can be build the element from top to bottom
        it is possible proof tha the operation cost less, but for now
        in python we mantains this approach.

        Time Complexity O(log N)
        """
        self.heap.append(value)
        self.__swap_parent(len(self.heap) - 1)
        return True

    def peek(self) -> int:
        if self.len() == 0:
            raise Exception("Empty data structure")
        return self.heap[0]

    def len(self) -> int:
        return len(self.heap)

    def to_list(self) -> [int]:
        return self.heap

    def verify(self) -> bool:
        for currentIdx in range(1, len(self.heap)):
            parent_idx = Heap.parent(currentIdx)
            if not self.cmp(parent_idx, currentIdx):
                return False
        return True

    def heapify(self, node, size_offset=0):
        """Restore the propriety inside the heap"""
        left_node = Heap.left(node)
        right_node = Heap.right(node)
        target_node = node
        if (self.len() - 1 - size_offset >= left_node) and (
            not self.cmp(target_node, left_node)
        ):
            target_node = left_node
        elif (self.len() - 1 - size_offset >= right_node) and (
            not self.cmp(target_node, right_node)
        ):
            target_node = right_node

        if target_node != node:
            self.swap(node, target_node)
            self.heapify(target_node)

    def __swap_parent(self, node: int) -> None:
        """Take a node and check if it is possible swap it with the parent
        this mean that the heap violate the heap propriety"""
        if node is None or node < 0:
            return
        parent_idx = Heap.parent(node)
        if not self.cmp(parent_idx, node):
            self.swap(node, parent_idx)
            self.__swap_parent(parent_idx)

    def swap(self, idx_one: int, idx_two: int) -> None:
        """Swap the position of the two element in index {idx_two} and {idx_two}"""
        self.heap[idx_one], self.heap[idx_two] = self.heap[idx_two], self.heap[idx_one]
        # print(f"Swapping {self.heap[idx_one]} <-> {self.heap[idx_two]}")

    def __str__(self):
        result = ""
        for _, value in enumerate(self.heap):
            result += f"{value}, "
        return result

    def __sizeof__(self):
        return self.len()

    @abstractmethod
    def cmp(self, idx_one: int, idx_two: int) -> bool:
        """Compare the two element in position {idx_one} and {idx_two}"""
        pass


class HeapTopDown(ABC):
    """Optimization of the Heap datastructure where the construction time move from
    O(N log N) to O(N), by the moving of the shift down of the element.
    FIXME: this is done also from the previous Heap implementation from the heap data structure, right?"""

    def __init__(self, array=[]):
        self.heap = array
        self.heap = self.from_array(array)

    @staticmethod
    def parent(idx: int) -> int:
        return (idx - 1) // 2

    @staticmethod
    def left(idx: int) -> int:
        return (idx * 2) + 1

    @staticmethod
    def right(idx: int) -> int:
        return (idx * 2) + 2

    def from_array(self, array) -> list:
        parent_id = self.parent(len(array) - 1)
        for currentIdx in reversed(range(parent_id + 1)):
            self.sift_down(currentIdx, len(array) - 1)
        return self.heap

    def sift_down(self, start_idx, end_idx):
        left_idx = self.left(start_idx)
        while left_idx <= end_idx:
            right_idx = self.right(start_idx)
            idx_to_swap = left_idx
            if right_idx <= end_idx:
                if self.cmp(right_idx, left_idx):
                    idx_to_swap = right_idx

            if self.cmp(idx_to_swap, start_idx):
                self.swap(start_idx, idx_to_swap)
                start_idx = idx_to_swap
                left_idx = self.left(start_idx)
            else:
                break

    def sift_up(self, start_idx):
        """Bring the node to the top of the tree"""
        parent_idx = self.parent(start_idx)
        while parent_idx > 0 and self.cmp(start_idx, parent_idx):
            self.swap(start_idx, parent_idx)
            start_idx = parent_idx
            parent_idx = self.parent(start_idx)

    def peek(self):
        if len(self.heap) == 0:
            raise Exception("Heap empty")
        return self.heap[0]

    def remove(self) -> int:
        value = self.peek()
        self.swap(0, len(self.heap) - 1)
        self.heap.pop()  # remove from the list the elem.
        self.sift_down(0, len(self.heap) - 1)
        return value

    def insert(self, value) -> None:
        self.heap.append(value)
        self.sift_up(len(self.heap) - 1)

    @abstractmethod
    def cmp(self, start_idx: int, with_idx: int) -> bool:
        pass

    def swap(self, first_idx, second_idx):
        self.heap[first_idx], self.heap[second_idx] = (
            self.heap[second_idx],
            self.heap[first_idx],
        )

    def verify(self) -> bool:
        for currentIdx in range(1, len(self.heap)):
            parent_idx = Heap.parent(currentIdx)
            if not self.cmp(parent_idx, currentIdx):
                return False
        return True

    def len(self) -> int:
        return len(self.heap)

    def to_list(self) -> [int]:
        return self.heap

    def __sizeof__(self):
        return self.len()

    def __str__(self):
        accumulate = ""
        for _, elem in enumerate(self.heap):
            accumulate += f"{elem}, "
        return accumulate



from cpstl.datatype.heap import Heap, HeapTopDown


class MinHeap(Heap):
    """Min heap Python 3 implementation"""

    def cmp(self, idx_one: int, idx_two: int) -> bool:
        return self.heap[idx_one] <= self.heap[idx_two]

    def is_min_heap(self) -> bool:
        print(f"Min heap: {self}")
        return super().verify()


class MinHeapTopDown(HeapTopDown):
    """Use optimization of heap to improve the insert time from O(N log N) to O(N)"""

    def cmp(self, start_idx: int, with_idx: int) -> bool:
        start_node = self.heap[start_idx]
        with_node = self.heap[with_idx]
        return start_node <= with_node

    def is_min_heap(self) -> bool:
        print(f"Min heap: {self}")
        return super().verify()



from abc import ABC, abstractmethod
from enum import Enum


class Order(Enum):
    INCREASE = 1
    DECREASE = 2


class Sort(ABC):
    """Sort Generic interface that implement all the method that
    a sort algorithm need to implement"""

    @abstractmethod
    def sort(self, sequence: list) -> list:
        """Take in input the sequence to sort and retyrn
        as result the same sequence sorted"""
        pass



"""
Jaccard Sim algorithm implementation in pure python 3.
Copyright (C) 2021-2022 Vincenzo Palazzo vincenzopalazzodev@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
USA.
"""
from .interface import ISimilarity


class Jaccard(ISimilarity):
    """
    Jaccard Sim algorithm implementation in pure python 3.

    author: Vincenzo Palazzo https://github.com/vincenzopalazzo
    """

    @staticmethod
    def __intersection_set(set_a: set, set_b: set) -> set:
        """Make the intersection of the two lists"""
        return set(set_a).intersection(set_b)

    @staticmethod
    def __union_set(set_a: set, set_b: set) -> set:
        """Make the union of the two list"""
        return set(set_a).union(set_b)

    def similarity(self, set_a: set, set_b: set) -> set:
        intersection = Jaccard.__intersection_set(set_a, set_b)
        union = Jaccard.__union_set(set_a, set_b)
        result = set()
        for _, elem in enumerate(union):
            if elem in intersection:
                result.add(elem)
        return result



"""
Universal Hashing algorithm in pure python 3.
Copyright (C) 2021-2022 Vincenzo Palazzo vincenzopalazzodev@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
USA.
"""
import random
from .utils import Utils


class UniversalHash:
    """
    Universal Hashing algorithm in pure python 3.

    author: Vincenzo Palazzo https://github.com/vincenzopalazzo
    """

    def __init__(self, size: int = -1):
        """
        Init the Universal Hash with a default size of 10 (take a random number, no motivation here)

        :param size: is equivalent in the Computer science theory to the Universe of values.
        """
        self.value_a = -1
        self.value_b = -1
        self.size = size
        self.prime_number = Utils.generate_random_prime(size)
        self.random_choice()

    def random_choice(self) -> None:
        """Make a random choice"""
        self.value_a = random.randint(1, self.prime_number - 1)
        self.value_b = random.randint(0, self.prime_number - 1)

    def hash(self, value_to_hash: int) -> int:
        """Calculate the universal hash value of the value"""
        assert self.value_a >= 0 and self.value_b >= 0, (
            "Random A and B value are not initialized, this should " "never happens "
        )
        return (
            (self.value_a * value_to_hash * self.value_b) % self.prime_number
        ) % self.size

    def hash_a_str(self, value_to_hash: str) -> int:
        """Calculate the Universal Hash function of the string value"""
        accumulator = 0
        binary_str = Utils.convert_str_to_byte(value_to_hash)
        for elem in binary_str:
            accumulator += int(elem)
        return self.hash(accumulator)



"""
Rolling Hash algorithm in pure python 3.
Copyright (C) 2021-2022 Vincenzo Palazzo vincenzopalazzodev@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
USA.
"""
import random
from .utils import Utils


class RollingHash:
    """
    Karp Rabin Algorithm implementation in pure python 3
    with reference implementation of the C++ implementation.

    Reference implementation: https://github.com/vincenzopalazzo/cpstl/blob/master/cpp/algoritmins/karp_rabin_fingerprint/KarpRabinHash.hpp
    """

    def __init__(self) -> None:
        # FIXME: remove this static array, and add an
        # and algorithm to choose the prime number.
        self.prime_numbers = [
            62563,
            62581,
            62591,
            62597,
            62603,
            62617,
            62627,
            62633,
            62639,
            62653,
        ]
        self.prime_number = random.choice(self.prime_numbers)

    def hash_str(self, str_to_hash: str) -> int:
        """Compute the hash of the input string"""
        return self.hash_from_byte(Utils.convert_str_to_byte(str_to_hash))

    def hash_from_byte(self, sequence_of_byte: str) -> int:
        """Compute the hash of the sequence of byte given in input"""
        fingerprint = 0
        for char in sequence_of_byte:
            bit = int(char)
            # x' = ((x « 1) + bit[i]) % p
            fingerprint = ((fingerprint * 2) + bit) % self.prime_number
        return fingerprint



"""
Count Min Sketch algorithm in pure python 3.
Copyright (C) 2021-2022 Vincenzo Palazzo vincenzopalazzodev@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
USA.
"""
from array import array
from math import pow, log
from ..hash.universal_hash import UniversalHash


def generate_n_hash_function(
    number_hashes: int, universe_hash: int = int(pow(2, 32))
) -> list:
    """
    Generate N hash functions!
    :param universe_hash: The size of the universe of the hash functions
    :param number_hashes: Number of hash functions!
    :return: return the list of hash function
    """
    hashes = []
    for _ in range(number_hashes):
        hashes.append(UniversalHash(size=universe_hash))
    return hashes


class CountMinSketch:
    """
    Count Min Sketch algorithm in pure python 3.

    The Count min sketch will divide the algorithm in different steps, like:
    - Step 1: Build a Matrix of size [C x R] with all possession sets to zero
    - Where the Colum is calculated with the formula `C = eulero_contant / approximation`
    and the row is calculated with the formula `R = ln(1 / gamma)`

    author: Vincenzo Palazzo https://github.com/vincenzopalazzo
    """

    def __init__(self, approximation: float = 0.001):
        self.size = 0
        self.matrix = []
        colum = CountMinSketch.calculate_colums(approximation)
        row = CountMinSketch.calculate_rows()
        for _ in range(row):
            table = array("l", (0 for _ in range(colum)))
            self.matrix.append(table)

        # generate R hash functions, with a universe that is `h -> [c]`
        self.hash_functions = generate_n_hash_function(
            number_hashes=row, universe_hash=colum
        )

    @staticmethod
    def calculate_colums(approximation: float) -> int:
        """
        Calculate the colums size from an approximation number
        :param approximation:
        :return: Size of the colums calculate with the formula `C = eulero_contant / approximation`
        """
        eulero_constant = 2.71828
        return int(eulero_constant / approximation)

    @staticmethod
    def calculate_rows() -> int:
        """"""
        return int(log((1 / (1 / 3)), 2))

    def __compute_hashes(self, val):
        if not isinstance(val, int):
            raise Exception(
                "Value not an instance of int, other type are not supported by the Universal Hash function!"
            )
        for idx in range(len(self.hash_functions)):
            yield self.hash_functions[idx].hash(val)

    def increment(self, value, increment: int = 1) -> None:
        """Apply an increment to the value"""
        self.size += increment
        for table, i in zip(self.matrix, self.__compute_hashes(value)):
            table[i] += increment

    def query(self, value) -> int:
        """Query the data structure and return the estimation frequency of the value"""
        print(self.matrix)
        return min(
            table[i] for table, i in zip(self.matrix, self.__compute_hashes(value))
        )

    def __getitem__(self, item):
        return self.query(item)

    def __len__(self):
        return self.size



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cpstl',
 'cpstl.algorithm',
 'cpstl.algorithm.hash',
 'cpstl.algorithm.similarity',
 'cpstl.algorithm.sketch',
 'cpstl.algorithm.sort',
 'cpstl.datatype',
 'cpstl.datatype.graph',
 'cpstl.datatype.heap']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'cpstl',
    'version': '0.0.5',
    'description': 'Copy and Paste standard library (CPSTL) is a repository with a collection of data structure and algorithms in many different languages',
    'long_description': '# Python Copy and Paste standard Library\n\nPython Copy and Paste standard library (CPSTL) is a repository with a collection of data structure and algorithms.\n\n## TODO\nAdd more content\n\n\n## License\n\n<div align="center">\n  <img src="https://opensource.org/files/osi_keyhole_300X300_90ppi_0.png" width="150" height="150"/>\n</div>\n\n    Collection of algorithm\'s and data struct developer in different language\n    and with the Code Style guide line, to resolve Competitive Programming problem.\n    Copyright (C) 2020-2021 Vincenzo Palazzo vincenzopalazzodev@gmail.com\n\n    This program is free software; you can redistribute it and/or modify\n    it under the terms of the GNU General Public License as published by\n    the Free Software Foundation; either version 2 of the License, or\n    (at your option) any later version.\n\n    This program is distributed in the hope that it will be useful,\n    but WITHOUT ANY WARRANTY; without even the implied warranty of\n    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n    GNU General Public License for more details.\n\n    You should have received a copy of the GNU General Public License along\n    with this program; if not, write to the Free Software Foundation, Inc.,\n    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.',
    'author': 'Vincenzo Palazzo',
    'author_email': 'vincenzopalazzodev@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://vincenzopalazzo.github.io/cpstl',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from .sort import Sort, Order, HeapSort
from .hash import UniversalHash
from .sketch.count_min_sketch import CountMinSketch
from .sketch.flajolet_martin import FMSketch
from .similarity import MinHash, Jaccard, ISimilarity



"""
Min Hash algorithm implementation in pure python 3.
Copyright (C) 2021-2022 Vincenzo Palazzo vincenzopalazzodev@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
USA.
"""
from typing import Any

from .interface import ISimilarity
from .jaccard_similarity import Jaccard
from cpstl.algorithm.hash import UniversalHash
from cpstl.datatype.heap import MinHeap


class MinHash(ISimilarity):
    """
    Min Hash algorithm implementation in pure python 3.

    The algorithm is divided in different steps:
    1. Generate K random function (we use the universal hash function here);
    2. Generate the signature of each value for each hash function and take the minimum;
    3.

    author: Vincenzo Palazzo https://github.com/vincenzopalazzo
    """

    def __init__(self, hash_function_size: int, hash_function_universe: int) -> None:
        """TODO: adding the theoretical estimator for the hash_function_size"""
        self.hash_function = MinHash.__generate_hash_functions(
            hash_function_size, hash_function_universe
        )
        self.hash_to_real_value_map = dict()

    @staticmethod
    def __generate_hash_functions(size: int, universe: int) -> list:
        """Generate hash function from the input parameter

        :param size: How much hash functions need to be generated.
        :param universe: Universe of value of the hash function.
        :return list of generate hash function
        """
        hash_function = []
        for _ in range(size):
            fn = UniversalHash(size=universe)
            hash_function.append(fn)
        return hash_function

    def __add_to_map_if_missed(self, element_value: Any, element_hash: int) -> None:
        """
        Utils function that adds to a map the real value that the hash corresponds with.
        """
        if element_hash not in self.hash_to_real_value_map:
            self.hash_to_real_value_map[element_hash] = element_value

    def __take_minimum_hash_value(self, value: Any) -> int:
        """
        Use the min heap to return the minimum value of the hash functions
        :return: Minimum value between hash function result.
        """
        # due a bug inside the heap insert function
        # we need to use this work around for now
        hash_min_heap = []

        for _, fn in enumerate(self.hash_function):
            hash_elem = fn.hash_a_str(str(value))
            hash_min_heap.append(int(hash_elem))
        return MinHeap(array=hash_min_heap).peek()

    def similarity(self, set_a: list, set_b: list) -> list:
        """
        Calculate the Min Hash similarity between the two sets
        :param set_a: first set of elements
        :param set_b: second set of elements
        :return: the list of common element between the two sets
        """
        hash_set_a = set()
        hash_set_b = set()

        # feel the hash_set_a
        for _, value in enumerate(set_a):
            min_hash = self.__take_minimum_hash_value(value)
            self.__add_to_map_if_missed(element_value=value, element_hash=min_hash)
            hash_set_a.add(min_hash)

        # feel the hash_set_b
        for _, value in enumerate(set_b):
            min_hash = self.__take_minimum_hash_value(value)
            self.__add_to_map_if_missed(element_value=value, element_hash=min_hash)
            hash_set_b.add(min_hash)

        common = Jaccard().similarity(hash_set_a, hash_set_b)
        result = []
        for _, comm_elem in enumerate(common):
            if comm_elem in self.hash_to_real_value_map:
                real_value = self.hash_to_real_value_map[comm_elem]
                result.append(real_value)
        return result



"""
Graph data structure implementation with adjacency list in Python 3
Copyright (C) 2021 Vincenzo Palazzo vincenzopalazzodev@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
USA.
"""
# FIXME: implementing it in the package and use it
from queue import Queue
from .graph import Node, Graph


class GraphList(Graph):
    """
    Class implementation of graph with adjacency list.
    This class use contains a list of node
    """

    def __init__(self, directed=False) -> None:
        self.directed = directed
        self.nodes = []

    def __dfs_helper(self, target: Node, path: list) -> None:
        path.append(target.value)
        for node in target.children:
            self.__dfs_helper(node, path)

    def __bsf_helper(self, target: Node, path: list) -> None:
        to_visit = Queue()
        to_visit.put(target)
        while not to_visit.empty():
            node = to_visit.get()
            path.append(node.value)
            for child in node.children:
                to_visit.put(child)

    def __find_node(self, node: Node, value) -> Node:
        """
        Find the node with the value in the node
        if it exist return the node, otherwise return None
        """
        if node.value == value:
            return node
        for child in node.children:
            found = self.__find_node(child, value)
            if found is not None:
                return found
        return None

    def __find_node_add(self, node: Node, value, to_add) -> bool:
        """
        Find the node with the value in the node
        if it exist return the node, otherwise return None
        """
        if node.value == value:
            node.add_node(to_add)
            return True
        for child in node.children:
            found = self.__find_node(child, value)
            if found is True:
                return True
        return False

    def add_edge(self, u, v, opts: dict = None) -> None:
        """
        Adding an edge to graph, if directed is true
        add the edge from u -> v, otherwise adding an
        edge u -> v and v -> u.
        """
        if self.directed:
            if len(self.nodes) != 0:
                for node in self.nodes:
                    found = self.__find_node_add(node, u, v)
                    if found is True:
                        return
            node = Node(u)
            node.add_node(v)
            self.nodes.append(node)
        else:
            raise RuntimeError("unsupported undirected graph")

    def dfs(self) -> list:
        paths = []
        for node in self.nodes:
            path = []
            self.__dfs_helper(node, path)
            paths.append(path)
        return paths

    def bfs(self) -> list:
        paths = []
        for node in self.nodes:
            path = []
            self.__bsf_helper(node, path)
            print(path)
            paths.append(path)
        return paths

    def size(self) -> int:
        uniques = set()
        for node in self.nodes:
            uniques.add(node.value)
            for child in node.children:
                uniques.add(child.value)
        return len(uniques)
